import { __awaiter, __generator } from "tslib";
import { Kernel, BufferUsage } from '@antv/g-plugin-gpgpu';
import { convertGraphData2CSC } from '../util';
/**
 * Pagerank using power method, ported from CUDA
 *
 * @param graphData
 * @param eps Set the tolerance the approximation, this parameter should be a small magnitude value. The lower the tolerance the better the approximation.
 * @param alpha The damping factor alpha represents the probability to follow an outgoing edge, standard value is 0.85.
 * @param maxIteration Set the maximum number of iterations.
 *
 * @see https://github.com/princeofpython/PageRank-with-CUDA/blob/main/parallel.cu
 */

export function pageRank(device, graphData, eps, alpha, maxIteration) {
  if (eps === void 0) {
    eps = 1e-05;
  }

  if (alpha === void 0) {
    alpha = 0.85;
  }

  if (maxIteration === void 0) {
    maxIteration = 1000;
  }

  return __awaiter(this, void 0, void 0, function () {
    var BLOCK_SIZE, BLOCKS, _a, V, From, To, n, graph, r, j, sum, i, i, storeKernel, matmulKernel, rankDiffKernel, rBuffer, rLastBuffer, graphBuffer, readback, grids, last, result, out;

    return __generator(this, function (_b) {
      switch (_b.label) {
        case 0:
          BLOCK_SIZE = 1;
          BLOCKS = 256;
          _a = convertGraphData2CSC(graphData), V = _a.V, From = _a.From, To = _a.To;
          n = V.length;
          graph = new Float32Array(new Array(n * n).fill((1 - alpha) / n));
          r = new Float32Array(new Array(n).fill(1 / n));
          From.forEach(function (from, i) {
            graph[To[i] * n + from] += alpha * 1.0;
          });

          for (j = 0; j < n; j++) {
            sum = 0.0;

            for (i = 0; i < n; ++i) {
              sum += graph[i * n + j];
            }

            for (i = 0; i < n; ++i) {
              if (sum != 0.0) {
                graph[i * n + j] /= sum;
              } else {
                graph[i * n + j] = 1 / n;
              }
            }
          }

          storeKernel = new Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<f32>;\n};\n\n@group(0) @binding(0) var<storage, read> r : Buffer;\n@group(0) @binding(1) var<storage, write> r_last : Buffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    r_last.data[index] = r.data[index];\n  }\n}")
          });
          matmulKernel = new Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<f32>;\n};\n\n@group(0) @binding(0) var<storage, read> graph : Buffer;\n@group(0) @binding(1) var<storage, read_write> r : Buffer;\n@group(0) @binding(2) var<storage, read> r_last : Buffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    var sum = 0.0;\n    for (var i = 0u; i < ").concat(V.length, "u; i = i + 1u) {\n      sum = sum + r_last.data[i] * graph.data[index * ").concat(V.length, "u + i];\n    }\n    r.data[index] = sum;\n  }\n}\n    ")
          });
          rankDiffKernel = new Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<f32>;\n};\n\n@group(0) @binding(0) var<storage, read> r : Buffer;\n@group(0) @binding(1) var<storage, read_write> r_last : Buffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    r_last.data[index] = abs(r_last.data[index] - r.data[index]);\n  }\n}    \n    ")
          });
          rBuffer = device.createBuffer({
            usage: BufferUsage.STORAGE | BufferUsage.COPY_SRC,
            viewOrSize: new Float32Array(r)
          });
          rLastBuffer = device.createBuffer({
            usage: BufferUsage.STORAGE | BufferUsage.COPY_SRC,
            viewOrSize: new Float32Array(n)
          });
          graphBuffer = device.createBuffer({
            usage: BufferUsage.STORAGE,
            viewOrSize: new Float32Array(graph)
          });
          readback = device.createReadback();
          storeKernel.setBinding(0, rBuffer);
          storeKernel.setBinding(1, rLastBuffer);
          matmulKernel.setBinding(0, graphBuffer);
          matmulKernel.setBinding(1, rBuffer);
          matmulKernel.setBinding(2, rLastBuffer);
          rankDiffKernel.setBinding(0, rBuffer);
          rankDiffKernel.setBinding(1, rLastBuffer);
          grids = Math.ceil(V.length / (BLOCKS * BLOCK_SIZE));
          _b.label = 1;

        case 1:
          if (!maxIteration--) return [3
          /*break*/
          , 3];
          storeKernel.dispatch(grids, 1);
          matmulKernel.dispatch(grids, 1);
          rankDiffKernel.dispatch(grids, 1);
          return [4
          /*yield*/
          , readback.readBuffer(rLastBuffer)];

        case 2:
          last = _b.sent();
          result = last.reduce(function (prev, cur) {
            return prev + cur;
          }, 0);

          if (result < eps) {
            return [3
            /*break*/
            , 3];
          }

          return [3
          /*break*/
          , 1];

        case 3:
          return [4
          /*yield*/
          , readback.readBuffer(rBuffer)];

        case 4:
          out = _b.sent();
          return [2
          /*return*/
          , Array.from(out).map(function (score, index) {
            return {
              id: graphData.nodes[index].id,
              score: score
            };
          }).sort(function (a, b) {
            return b.score - a.score;
          })];
      }
    });
  });
}