"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.sssp = sssp;

var _tslib = require("tslib");

var _gPluginGpgpu = require("@antv/g-plugin-gpgpu");

var _util = require("../util");

/**
 * SSSP(Bellman-Ford) ported from CUDA
 *
 * @see https://www.lewuathe.com/illustration-of-distributed-bellman-ford-algorithm.html
 * @see https://github.com/sengorajkumar/gpu_graph_algorithms
 * @see https://docs.rapids.ai/api/cugraph/stable/api_docs/api/cugraph.traversal.sssp.sssp.html
 * compared with G6:
 * @see https://g6.antv.vision/zh/docs/api/Algorithm#findshortestpathgraphdata-start-end-directed-weightpropertyname
 */
function sssp(device, graphData, sourceId, weightPropertyName, maxDistance) {
  if (weightPropertyName === void 0) {
    weightPropertyName = '';
  }

  if (maxDistance === void 0) {
    maxDistance = 1000000;
  }

  return (0, _tslib.__awaiter)(this, void 0, void 0, function () {
    var BLOCK_SIZE, BLOCKS, MAX_DISTANCE, _a, V, E, I, nodeId2IndexMap, edges, W, sourceIdx, relaxKernel, updateDistanceKernel, updatePredKernel, VBuffer, EBuffer, IBuffer, WBuffer, view, DOutBuffer, DiOutBuffer, POutBuffer, readback, grids, i, out, predecessor;

    return (0, _tslib.__generator)(this, function (_b) {
      switch (_b.label) {
        case 0:
          BLOCK_SIZE = 1;
          BLOCKS = 256;
          MAX_DISTANCE = maxDistance;
          _a = (0, _util.convertGraphData2CSC)(graphData), V = _a.V, E = _a.E, I = _a.I, nodeId2IndexMap = _a.nodeId2IndexMap, edges = _a.edges;
          sourceIdx = nodeId2IndexMap[sourceId];

          if (weightPropertyName) {
            W = edges.map(function (edgeConfig) {
              return Number(edgeConfig[weightPropertyName]);
            });
          } else {
            // all the vertex has the same weight
            W = new Array(E.length).fill(1);
          }

          relaxKernel = new _gPluginGpgpu.Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<i32>;\n};\nstruct AtomicBuffer {\n  data: array<atomic<i32>>;\n};\n\n@group(0) @binding(0) var<storage, read> d_in_E : Buffer;\n@group(0) @binding(1) var<storage, read> d_in_I : Buffer;\n@group(0) @binding(2) var<storage, read> d_in_W : Buffer;\n@group(0) @binding(3) var<storage, read> d_out_D : Buffer;\n@group(0) @binding(4) var<storage, read_write> d_out_Di : AtomicBuffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    for (var j = d_in_I.data[index]; j < d_in_I.data[index + 1u]; j = j + 1) {\n      var w = d_in_W.data[j];\n      var du = d_out_D.data[index];\n      var dv = d_out_D.data[d_in_E.data[j]];\n      var newDist = du + w;\n      if (du == ").concat(MAX_DISTANCE, ") {\n        newDist = ").concat(MAX_DISTANCE, ";\n      }\n\n      if (newDist < dv) {\n        atomicMin(&d_out_Di.data[d_in_E.data[j]], newDist);\n      }\n    }\n  }\n}")
          });
          updateDistanceKernel = new _gPluginGpgpu.Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<i32>;\n};\n\n@group(0) @binding(0) var<storage, read_write> d_out_D : Buffer;\n@group(0) @binding(1) var<storage, read_write> d_out_Di : Buffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    if (d_out_D.data[index] > d_out_Di.data[index]) {\n      d_out_D.data[index] = d_out_Di.data[index];\n    }\n    d_out_Di.data[index] = d_out_D.data[index];\n  }\n}\n    ")
          });
          updatePredKernel = new _gPluginGpgpu.Kernel(device, {
            computeShader: "\nstruct Buffer {\n  data: array<i32>;\n};\nstruct AtomicBuffer {\n  data: array<atomic<i32>>;\n};\n\n@group(0) @binding(0) var<storage, read> d_in_V : Buffer;\n@group(0) @binding(1) var<storage, read> d_in_E : Buffer;\n@group(0) @binding(2) var<storage, read> d_in_I : Buffer;\n@group(0) @binding(3) var<storage, read> d_in_W : Buffer;\n@group(0) @binding(4) var<storage, read> d_out_D : Buffer;\n@group(0) @binding(5) var<storage, read_write> d_out_P : AtomicBuffer;\n\n@stage(compute) @workgroup_size(".concat(BLOCKS, ", ").concat(BLOCK_SIZE, ")\nfn main(\n  @builtin(global_invocation_id) global_id : vec3<u32>\n) {\n  var index = global_id.x;\n  if (index < ").concat(V.length, "u) {\n    for (var j = d_in_I.data[index]; j < d_in_I.data[index + 1u]; j = j + 1) {\n      var u = d_in_V.data[index];\n      var w = d_in_W.data[j];\n\n      var dis_u = d_out_D.data[index];\n      var dis_v = d_out_D.data[d_in_E.data[j]];\n      if (dis_v == dis_u + w) {\n        atomicMin(&d_out_P.data[d_in_E.data[j]], u);\n      }\n    }\n  }\n}    \n    ")
          });
          VBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE,
            viewOrSize: new Int32Array(V)
          });
          EBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE,
            viewOrSize: new Int32Array(E)
          });
          IBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE,
            viewOrSize: new Int32Array(I)
          });
          WBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE,
            viewOrSize: new Int32Array(W)
          });
          view = new Array(V.length).fill(MAX_DISTANCE);
          view[sourceIdx] = 0;
          DOutBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE | _gPluginGpgpu.BufferUsage.COPY_SRC,
            viewOrSize: new Int32Array(view)
          });
          DiOutBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE | _gPluginGpgpu.BufferUsage.COPY_SRC,
            viewOrSize: new Int32Array(view)
          });
          POutBuffer = device.createBuffer({
            usage: _gPluginGpgpu.BufferUsage.STORAGE | _gPluginGpgpu.BufferUsage.COPY_SRC,
            viewOrSize: new Int32Array(view)
          });
          readback = device.createReadback();
          relaxKernel.setBinding(0, EBuffer);
          relaxKernel.setBinding(1, IBuffer);
          relaxKernel.setBinding(2, WBuffer);
          relaxKernel.setBinding(3, DOutBuffer);
          relaxKernel.setBinding(4, DiOutBuffer);
          updateDistanceKernel.setBinding(0, DOutBuffer);
          updateDistanceKernel.setBinding(1, DiOutBuffer);
          updatePredKernel.setBinding(0, VBuffer);
          updatePredKernel.setBinding(1, EBuffer);
          updatePredKernel.setBinding(2, IBuffer);
          updatePredKernel.setBinding(3, WBuffer);
          updatePredKernel.setBinding(4, DOutBuffer);
          updatePredKernel.setBinding(5, POutBuffer);
          grids = Math.ceil(V.length / (BLOCKS * BLOCK_SIZE));

          for (i = 1; i < V.length; i++) {
            relaxKernel.dispatch(grids, 1);
            updateDistanceKernel.dispatch(grids, 1);
          }

          updatePredKernel.dispatch(grids, 1);
          return [4
          /*yield*/
          , readback.readBuffer(DiOutBuffer)];

        case 1:
          out = _b.sent();
          return [4
          /*yield*/
          , readback.readBuffer(POutBuffer)];

        case 2:
          predecessor = _b.sent();
          return [2
          /*return*/
          , Array.from(out).map(function (distance, i) {
            return {
              target: graphData.nodes[V[i]].id,
              distance: distance,
              predecessor: graphData.nodes[V[predecessor[i]]].id
            };
          })];
      }
    });
  });
}